import logging
import sys
import gemmi
import csv
import sqlite3
from contextlib import contextmanager
import requests
import jq
import pickle
import os
import json
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class CIFReader:
    def __init__(self) -> None:
        self.data = {}  # Dictionary to store the parsed CIF data
        

    def read_files(self, file_paths):
        logging.info("Reading CIF files")
        for file_path in file_paths:
            cif_block = gemmi.cif.read_file(file_path)
            file_name = file_path.split("/")[-1]
            self.data[file_name] = cif_block.sole_block()
    
    def item_exists_across_all(self, category, item):
        logging.info(f"Checking existence across all model files for {category}.{item}")
        try:
            for file_name, sole_block in self.data.items():
                if sole_block.find_mmcif_category(category):
                    if not sole_block.get_mmcif_category(category)[item]:
                        logging.info("Does not exist across all files")
                        return False
        except KeyError as e:
            logging.warning(f"Missing {category}.{item} from {file_name}")
            logging.info("Does not exist across all files")
            return False
        logging.info("Exists across all files")
        return True

    def item_is_empty_in_any(self, category, item):
        logging.info(f"Checking if any items is empty in {category}.{item}")
        for file_name, sole_block in self.data.items():
            if sole_block.find_mmcif_category(category):
                try:
                    for value in sole_block.get_mmcif_category(category)[item]:
                        if value in [None, "?"]:
                            logging.info(f"Value is empty in {file_name} for {category}.{item}")
                            return True
                except Exception as e:
                    logging.warning(f"Missing field {category}.{item} from {file_name}")
                    return True
        logging.info("Value is non empty in all files")
        return False

    def collate_item_per_file(self, category, item):
        logging.info(f"Collating with distinct files {category}.{item}")
        collated_data = {}
        for file_no, (file_name, sole_block) in enumerate(self.data.items()):
            collated_data[file_no] = []
            if sole_block.find_mmcif_category(category):
                for values in sole_block.get_mmcif_category(category)[item]:
                    collated_data[file_no].append(values)
        return collated_data

    def collate_item(self, category, item):
        logging.info(f"Collating {category}.{item}")
        collated_data = []
        try:
            for file_name, sole_block in self.data.items():
                if sole_block.find_mmcif_category(category):
                    for values in sole_block.get_mmcif_category(category)[item]:
                        collated_data.append(values)
        except KeyError as e:
            logging.exception(f"Missing {category}.{item} from {file_name}")
            raise Exception(e)
        return collated_data

    def collate_items(self, category, items):
        logging.info(f"Collating multiple items in {category} items: {items}")
        collated_data = {}
        for file_name, sole_block in self.data.items():
            try:
                if sole_block.find_mmcif_category(category):
                    for item in items:
                        collated_data.setdefault(item, [])
                        for values in sole_block.get_mmcif_category(category)[item]:
                            collated_data[item].append(values)
            except Exception as e:
                logging.exception(f"Missing {category}.{item} from {file_name}")
                raise Exception(e)
        return collated_data

    def collate_category(self, category):
        logging.info(f"Collating all items in {category}")
        collated_data = {}
        for file_name, sole_block in self.data.items():
            if sole_block.find_mmcif_category(category):
                for item, values in sole_block.get_mmcif_category(category).items():
                    collated_data.setdefault(item, [])
                    collated_data[item].extend(values)
        return collated_data

    def get_data(self, category, items):
        filtered_data = []
        for file_name, sole_block in self.data.items():
            filtered_data.append(sole_block.get_mmcif_category(category)[items])
        return filtered_data

    def get_rows_in_category(self, category):
        for file_name, sole_block in self.data.items():
            if sole_block.find_mmcif_category(category):
                items = sole_block.get_mmcif_category(category)
                first_item = items.keys()[0]
                return len(first_item)

class SqliteReader:
    def __init__(self, sqlite_path) -> None:
        self.data = {} 
        self.denormalised_data = []
        self.conn = sqlite3.connect(sqlite_path, uri=True)

    @contextmanager
    def sqlite_db_connection(self):
        logging.debug("Re-using In-memory DB connection")
        conn = self.conn
        try:
            yield conn
        finally:
            conn.commit()

    def sql_execute(self, query):
        logging.debug(f"Executing query: {query}")
        result = []
        with self.sqlite_db_connection() as conn:
            response = conn.execute(query)
            for row in response:
                result.append(row)
        return result
    
    def create_table_from_csv(self, csv_file):
        logging.info(f"Creating table from CSV file: {csv_file}")
        table_name = os.path.splitext(os.path.basename(csv_file))[0]

        # DROP TABLE IF EXISTS
        drop_table_query = f"DROP TABLE IF EXISTS \"{table_name}\""
        self.sql_execute(drop_table_query)
        
        with open(csv_file, 'r', newline='') as f:
            csv_reader = csv.reader(f)
            headers = next(csv_reader)
            
            # Read a sample of rows to determine column types
            sample_rows = []
            for _ in range(100):  # Read up to 100 rows as a sample
                try:
                    sample_rows.append(next(csv_reader))
                except StopIteration:
                    break
            
            column_types = self._determine_column_types(headers, sample_rows)
            
            # Create the table
            create_table_query = f"CREATE TABLE IF NOT EXISTS \"{table_name}\" ({', '.join([f'`{header}` {column_types[header]}' for header in headers])})"
            self.sql_execute(create_table_query)
            
            # Insert data
            insert_query = f"INSERT INTO \"{table_name}\" ({', '.join([f'`{header}`' for header in headers])}) VALUES ({', '.join(['?' for _ in headers])})"
            with self.sqlite_db_connection() as conn:
                cursor = conn.cursor()
                cursor.executemany(insert_query, sample_rows)
                
                # Continue inserting the rest of the data
                f.seek(0)
                next(csv_reader)  # Skip header
                for row in csv_reader:
                    if row not in sample_rows:
                        cursor.execute(insert_query, row)
                
            logging.info(f"Table '{table_name}' created and populated with data from {csv_file}")

    def _determine_column_types(self, headers, sample_rows):
        column_types = {}
        for header in headers:
            column_data = [row[headers.index(header)] for row in sample_rows if row[headers.index(header)]]
            if all(self._is_integer(value) for value in column_data):
                column_types[header] = 'INTEGER'
            elif all(self._is_float(value) for value in column_data):
                column_types[header] = 'REAL'
            else:
                column_types[header] = 'TEXT'
        return column_types

    @staticmethod
    def _is_integer(value):
        try:
            int(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def _is_float(value):
        try:
            float(value)
            return True
        except ValueError:
            return False

class InvestigationStorage:
    def __init__(self, investigation_id):
        self.data = {}
        self.mmcif_order = {}
        self.investigation_id = investigation_id

    def add_category(self, category_name):
        if category_name not in self.data:
            self.data[category_name] = {}

    def set_item(self, category_name, item_name, item_value):
        if category_name not in self.data:
            self.add_category(category_name)
        self.data[category_name][item_name] = item_value

    def get_category_data(self, category):
        return self.data[category]

    def get_item_data(self, category, item):
        return self.data[category][item]

    def get_items_data(self, category, items):
        result = {}
        for item in items:
            result[item] = self.data[category][item]
        return result

    def set_items(self, category_name, data):
        if category_name not in self.data:
            self.add_category(category_name)
        for item, values in data.items():
            if item not in self.data[category_name]:
                self.data[category_name][item] = []
            for value in values:
                self.data[category_name][item].append(value)

    def get_data(self) -> dict:
        return self.data
    
    def get_item_order(self, category) -> list:
        return self.mmcif_order.get(category, [])

    def write_data_to_cif(self, output_file) -> None:
        logging.info("Writing Investigation cif file")
        write_options = gemmi.cif.WriteOptions()
        write_options.align_loops = 50
        write_options.align_pairs = 50

        doc = gemmi.cif.Document()
        block = doc.add_new_block(f"{self.investigation_id}")
        for category, items in self.data.items():
            ordered_category = {}
            ordered_items = self.get_item_order(category)
            for ordered_item in ordered_items:
                if ordered_item in items:
                    ordered_category[ordered_item]  = items.pop(ordered_item)
            ordered_category.update(items)
            block.set_mmcif_category(category, ordered_category)
        block.write_file(output_file, write_options)

    def integrity_check(self):
        inconsistent_keys = {}
        for dictionary_key, dictionary_values in self.data.items():
            max_length = max(len(values) for values in dictionary_values.values())
            inconsistent_lists = [
                (key, len(values))
                for key, values in dictionary_values.items()
                if len(values) != max_length
            ]
            if inconsistent_lists:
                inconsistent_keys[dictionary_key] = inconsistent_lists

        for dictionary_key, keys_lengths in inconsistent_keys.items():
            print(f"Dictionary '{dictionary_key}' has inconsistent list lengths:")
            for key, length in keys_lengths:
                print(f"   Key '{key}' has length {length}")

class PickleReader:
    def __init__(self, pickle_path):
        self.data = {}
        self.pickle_path = pickle_path
        self.load_pickle()

    def load_pickle(self):
        if self.pickle_path:
            with open(self.pickle_path, 'rb') as file:
                self.data = pickle.load(file)

class ExternalInformation:
    def __init__(self, filename) -> None:
        self.filename = filename
        self.inchi_keys = {}

    def _load_inchi_keys(self):
        if self.inchi_keys:
            return
        logging.info("Loading Inchikeys csv file")
        with open(self.filename, mode="r", newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                chem_comp_id = row["CHEM_COMP_ID"]
                descriptor = row["DESCRIPTOR"]
                self.inchi_keys[chem_comp_id] = descriptor

    def _get_inchi_key(self, chem_comp_id):
        return self.inchi_keys.get(chem_comp_id)

    def get_inchi_key(self, chem_comp_id):
        self._load_inchi_keys()
        return self._get_inchi_key(chem_comp_id)
    
class JsonReader:
    def __init__(self, json_path):
        self.data = {}
        self.json_path = json_path
        self.load_json()

    def load_json(self):
        with open(self.json_path, 'r') as file:
            self.data = json.load(file)
    
    def jq_filter(self, filter_string: str):
        """
        Apply a JQ filter to the JSON data
        
        Args:
            filter_string (str): The JQ filter to apply
            
        Returns:
            The filtered data - could be a single value, list, or dict depending on the filter
        """
        try:
            return jq.compile(filter_string).input(self.data).first()
        except Exception as e:
            logging.error(f"Failed to apply JQ filter: {filter_string}")
            logging.exception(e)
            raise

class RestReader:
    def __init__(self, base_url, username, password):
        self.base_url = base_url
        self.session = requests.Session()
        self.access_token = self.get_auth_token(username, password)
        self.session.headers.update({'Authorization': f'Bearer {self.access_token}'})

    def get_auth_token(self, username, password):
        url = f"{self.base_url}/auth"
        payload = {"grant_type": "password", 'username': username, 'password': password}
        response = self.session.post(url, json=payload)
        response.raise_for_status()
        return response.json()['access_token']

    def get(self, endpoint, params=None, filter_query=None):
        url = f"{self.base_url}{endpoint}"
        response = self.session.get(url, params=params)
        response.raise_for_status()
        return self.filter_response(response.json(), filter_query)

    def post(self, endpoint, data=None, filter_query=None):
        url = f"{self.base_url}{endpoint}"
        response = self.session.post(url, json=data)
        response.raise_for_status()
        return self.filter_response(response.json(), filter_query)

    def filter_response(self, response_json, filter_query):
        if filter_query:
            return jq.compile(filter_query).input(response_json).first()
        return response_json


import gemmi
from pathlib import Path

# Block.write_category behaves differently than category.append_row (with regards to quotes)
# gemmi.cif.quote adds quotes to strings where necessary
# It also adds quotes to the "?"
# gemmi.cif.quote_list quotes a list of string. 
# Unlike the quote function, it can take a None value, and return unquoted "?"



test_cif = Path('./data/test_gemmi.cif')
gemmi_cif = gemmi.cif.read(str(test_cif))

for block_a in gemmi_cif:
    category = block_a.find_mmcif_category("_audit_author")
    # columns = {name: i for i, name in enumerate(category.tags)}
    # print("Columns: "+str(columns))
    # category_len = len(category)
    one_value = gemmi.cif.quote("Name something")
    category.append_row([one_value, "4"])
    # The output does not have quotes, w
write_options = gemmi.cif.WriteOptions()
write_options.align_loops = 50
write_options.align_pairs = 50





block = gemmi_cif.sole_block()
loop = block.find_loop("_audit_author")
item_loop = block.find_loop_item("_audit_author.name")


import pdb
pdb.set_trace()
gemmi_cif.write_file("./data/test_gemmi_out.cif", write_options)





import unittest
from unittest.mock import MagicMock, patch
from mmcif_gen.operations import (
    UnionOperation,
    UnionDistinctOperation,
    AutoIncrementOperation,
    StaticValueOperation,
    ModifyOperation,
    IntersectionOperation,
    ConditionalUnionOperation,
    ConditionalDistinctUnionOperation,
    CopyFillOperation,
    CopyConditionalModificationOperation,
    CopyOperation,
    CopyForEachRowOperation,
    DeletionOperation,
    ExternalInformationOperation,
    SQLOperation,
    NoopOperation,
    ItemDoNotExist
)
from investigation_io import InvestigationStorage, CIFReader
import os


class TestOperations(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dummy1_cif_path = 'test/data/dummy1.cif'
        cls.dummy2_cif_path = 'test/data/dummy2.cif'

    def setUp(self):
        self.investigation_storage = InvestigationStorage("test_investigation")
        self.reader = CIFReader()
        self.mock_storage = MagicMock(spec=InvestigationStorage)
        self.mock_reader = CIFReader()
        self.mock_reader.read_files([self.dummy1_cif_path, self.dummy2_cif_path])
        # Ensure mock_storage has a data attribute that is a dictionary
        self.mock_storage.data = {}

    def test_union_operation(self):
        operation = UnionOperation(self.mock_storage, self.mock_reader)
        operation_data = {
            "source_category": "_audit_author",
            "source_items": ["name", "pdbx_ordinal"],
            "target_category": "_target",
            "target_items": ["name", "pdbx_ordinal"],
            "operation_parameters": {"primary_parameters": ["name"], "secondary_parameters": ["pdbx_ordinal"]}
        }
        operation.perform_operation(operation_data)
        self.mock_storage.set_items.assert_called_with(
            "_target", 
            {"name": ["Snee, M.", "Talon, R.", "Fowley, D.", "Name something","Snee, M.", "Talon, R.", "Fowley, D."], 
             "pdbx_ordinal": ["1", "2", "3", "4","1", "2", "3"]}
        )

    def test_union_distinct_operation(self):
        operation = UnionDistinctOperation(self.mock_storage, self.mock_reader)
        operation_data = {
            "source_category": "_audit_author",
            "source_items": ["name", "pdbx_ordinal"],
            "target_category": "_target",
            "target_items": ["name", "pdbx_ordinal"],
            "operation_parameters": {"primary_parameters": ["name"], "secondary_parameters": ["pdbx_ordinal"]}
        }
        operation.perform_operation(operation_data)
        self.mock_storage.set_items.assert_called_with(
            "_target", 
            {"name": ["Snee, M.", "Talon, R.", "Fowley, D.", "Name something"], 
             "pdbx_ordinal": ["1", "2", "3", "4"]}
        )

    def test_auto_increment_operation(self):
        operation = AutoIncrementOperation(self.mock_storage, self.mock_reader)
        self.mock_storage.data = {"category": {"item": ["data1", "data2"]}}
        operation_data = {
            "target_category": "category",
            "target_item": "incremented_item"
        }
        operation.perform_operation(operation_data)
        self.assertEqual(self.mock_storage.data["category"]["incremented_item"], [1, 2])

    def test_static_value_operation(self):
        operation = StaticValueOperation(self.mock_storage, self.mock_reader)
        self.mock_storage.data = {"category": {}}
        operation_data = {
            "target_category": "category",
            "target_items": ["item1", "item2"],
            "target_values": ["value1", "value2"]
        }
        operation.perform_operation(operation_data)
        self.assertEqual(self.mock_storage.data["category"]["item1"], ["value1"])
        self.assertEqual(self.mock_storage.data["category"]["item2"], ["value2"])

    def test_modify_operation(self):
        operation = ModifyOperation(self.mock_storage, self.mock_reader)
        # self.mock_reader.collate_item.return_value = ["data"]
        operation_data = {
            "source_category": "_struct",
            "source_items": ["entry_id"],
            "target_category": "target",
            "target_items": ["doi"],
            "operation_parameters": "pdb_{}"
        }
        self.mock_storage.data["target"] = {}
        operation.perform_operation(operation_data)
        self.mock_storage.add_category.assert_called_with("target")
        self.assertEqual(self.mock_storage.data["target"]["doi"], ["pdb_1CBS"])

    @unittest.skip("need to verify whether failure is an implementaiton issue")
    def test_intersection_operation(self):
        operation = IntersectionOperation(self.mock_storage, self.mock_reader)
        operation_data = {
            "source_category": "_audit_author",
            "source_items": ["name"],
            "target_category": "target",
            "target_items": ["different_author"]
        }
        self.mock_storage.data["target"] = {}
        operation.perform_operation(operation_data)
        self.mock_storage.add_category.assert_called_with("target")
        self.assertEqual(self.mock_storage.data["target"]["different_author"], ["Name something"])

    def test_conditional_union_operation(self):
        operation = ConditionalUnionOperation(self.mock_storage, self.mock_reader)
        operation_data = {
            "source_category": "_audit_author",
            "source_items": ["name"],
            "target_category": "target",
            "target_items": ["name"],
            "operation_parameters": {
                "conditional_variable": "_audit_author.pdbx_ordinal",
                "value": ["4"]
            }
        }
        operation.perform_operation(operation_data)
        self.mock_storage.set_items.assert_called_with("target", {"name": ["Name something"]})

    def test_conditional_distinct_union_operation(self):
        operation = ConditionalDistinctUnionOperation(self.mock_storage, self.mock_reader)
        operation_data = {
            "source_category": "_audit_author",
            "source_items": ["name"],
            "target_category": "target",
            "target_items": ["name"],
            "operation_parameters": {
                "conditional_variable": "_audit_author.pdbx_ordinal",
                "value": ["1"],
                "distinct_key": ["name"]
            }
        }
        operation.perform_operation(operation_data)
        self.mock_storage.set_items.assert_called_with("target", {"name": ["Snee, M."]})

    def test_copy_fill_operation(self):
        operation = CopyFillOperation(self.investigation_storage, self.mock_reader)
        self.investigation_storage.data = {"target": {"existing_item": ["row1", "row2"]},
                                  "source": {"copy_me": ["1cbs"]}}
        operation_data = {
            "source_category": "source",
            "source_items": ["copy_me"],
            "target_category": "target",
            "target_items": ["code"]
        }
        operation.perform_operation(operation_data)
        self.assertEqual(self.investigation_storage.data["target"]["code"], ["1cbs", "1cbs"])

    def test_copy_conditional_modification_operation(self):
        operation = CopyConditionalModificationOperation(self.investigation_storage, self.mock_reader)
        self.investigation_storage.data = {
            "target": {},
            "source": { "item": ["data1", "data2"],
                        "condition": ["True", "False"]}
        }
        operation_data = {
            "source_category": "source",
            "source_items": ["item"],
            "target_category": "target",
            "target_items": ["item"],
            "operation_parameters": {
                "conditional_variable": "source.condition",
                "value": "True",
                "modification": "modified_{}"
            }
        }
        operation.perform_operation(operation_data)
        self.assertEqual(self.investigation_storage.data["target"]["item"], ["modified_data1", "?"])

    def test_copy_operation(self):
        operation = CopyOperation(self.investigation_storage, self.mock_reader)
        self.investigation_storage.data = {"target": {},
                                           "source": {"item" : ["value1", "value2"]}}
        operation_data = {
            "source_category": "source",
            "source_items": ["item"],
            "target_category": "target",
            "target_items": ["item"]
        }
        operation.perform_operation(operation_data)
        self.assertEqual(self.investigation_storage.data["target"]["item"], ["value1", "value2"])

    def test_copy_for_each_row_operation(self):
        operation = CopyForEachRowOperation(self.investigation_storage, self.mock_reader)
        self.investigation_storage.data = {"target": {"existing_ordinal": ["1", "3"]}}
        
        operation_data = {
            "source_category": "_audit_author",
            "source_items": ["name"],
            "target_category": "target",
            "target_items": ["name"],
            "operation_parameters": {
                "conditional_variable": "_audit_author.pdbx_ordinal",
                "value": "{target.existing_ordinal}"
            }
        }
        operation.perform_operation(operation_data)
        self.assertEqual(self.investigation_storage.data["target"]["name"], ["Snee, M.", 'Fowley, D.'])

    def test_deletion_operation(self):
        operation = DeletionOperation(self.mock_storage, self.mock_reader)
        self.mock_storage.data = {"target": {"item": ["data"]}}
        operation_data = {
            "target_category": "target",
            "target_items": ["item"]
        }
        operation.perform_operation(operation_data)
        self.assertNotIn("item", self.mock_storage.data["target"])

    @unittest.skip("Need to correct this unit test")
    def test_external_information_operation(self):
        operation = ExternalInformationOperation(self.mock_storage, self.mock_reader)
        self.mock_storage.data = {"source": {"item": ["data"]}, "target": {}}
        operation_data = {
            "source_category": "source",
            "source_items": ["item"],
            "target_category": "target",
            "target_items": ["item"],
            "operation_parameters": {"file": "external.json"}
        }
        with patch('operations.ExternalInformation') as MockExternal:
            mock_external = MockExternal.return_value
            mock_external.get_inchi_key.return_value = "inchi_key"
            operation.perform_operation(operation_data)
            self.assertEqual(self.mock_storage.data["target"]["item"], ["inchi_key"])

if __name__ == '__main__':
    unittest.main()



from mmcif_gen.investigation_engine import InvestigationEngine
from mmcif_gen.investigation_io import RestReader
from typing import List
import sys
import logging

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

class InvestigationESRF(InvestigationEngine):
        
    def __init__(self, user: str, pw:str, investigation_id: str, output_path: str) -> None:
        logging.info("Instantiating ESRF Investigation subclass")
        self.rest_reader = RestReader("https://htxlab.embl.org/", user, pw)
        self.rest_reader.get_auth_token()
        super().__init__(investigation_id, output_path)

    def pre_run(self) -> None:
        logging.info("Pre-running")
        super().pre_run()
    
def esrf_subparser(subparsers, parent_parser):
    parser_pdbe = subparsers.add_parser("esrf",help="Parameter requirements for investigation files from ESRF data", parents=[parent_parser])
    parser_pdbe.add_argument(
        "-f", 
        "--model-folder", help="Directory which contains model files"
    )



from abc import ABC, abstractmethod
from copy import deepcopy
from mmcif_gen.investigation_io import ExternalInformation, InvestigationStorage
import logging
import sys
from typing import Dict, List, TYPE_CHECKING

logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class ItemDoNotExist(Exception):
    pass

class operationBase(ABC):
    def __init__(
        self, investigation_storage: InvestigationStorage, reader) -> None:
        self.investigation_storage = investigation_storage
        self.reader = reader

    @abstractmethod
    def perform_operation(sel, operation_data: dict) -> None:
        pass

    def check_items_exist_across_all_files(self, category, items):
        for item in items:
            if not self.reader.item_exists_across_all(category, item):
                return False
            elif self.reader.item_is_empty_in_any(category, item):
                return False
        return True

    def get_number_of_rows_in_data(self, data):
        items = list(data.keys())
        if not items:
            return 0
        total_rows = len(data[items[0]])
        return total_rows

    def filter_rows_by_items(self, rows, keys):
        filtered_dict = {key: rows[key] for key in keys if key in rows}
        return filtered_dict

    def reduce_rows_by_distinct_key(self, rows, distinct_keys):
        reduced_rows = {}
        distinct_values = set()
        items = list(rows.keys())
        total_rows = self.get_number_of_rows_in_data(rows)

        for index in range(total_rows):
            distinct_key = ""
            for key in distinct_keys:
                distinct_key = distinct_key + rows[key][index]
            if distinct_key in distinct_values:
                continue
            else:
                distinct_values.add(distinct_key)
                for item in items:
                    reduced_rows.setdefault(item, [])
                    reduced_rows[item].append(rows[item][index])
        return reduced_rows

    def remove_rows(self, rows, indices):
        indices.sort(reverse=True)
        items = list(rows.keys())
        total_rows = self.get_number_of_rows_in_data(rows)

        for index in indices:
            for item in items:
                del rows[item][index]
        return rows

    def evaluate_variable(self, value):
        if value[0] == "{":
            category, item = value[1:-1].split(".")
            return self.investigation_storage.data[category][item]
        else:
            return value

    def reduce_rows_by_condition(self, rows, condition_tag, condition_value):
        to_be_removed_indices = []
        condition_item = condition_tag.split(".")[-1]
        for index, item_value in enumerate(rows[condition_item]):
            if item_value not in condition_value:
                to_be_removed_indices.append(index)

        return self.remove_rows(rows, to_be_removed_indices)

    def rename_item(
        self, rows: Dict[str, List[str]], source: str, target: str
    ) -> Dict[str, List[str]]:
        rows[target] = rows.pop(source)
        return rows


class UnionOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing UnionOperation")

        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})

        distinct_keys = operation_parameters["primary_parameters"]
        key_exists = self.check_items_exist_across_all_files(
            source_category, distinct_keys
        )
        if not key_exists:
            distinct_keys = operation_parameters["secondary_parameters"]
        key_exists = self.check_items_exist_across_all_files(
            source_category, distinct_keys
        )
        if not key_exists:
            raise ItemDoNotExist

        collated_data = self.reader.collate_items(source_category, source_items)
        if target_items != "_same":
            for index, item in enumerate(target_items):
                self.rename_item(collated_data, source_items[index], item)
        self.investigation_storage.set_items(target_category, collated_data)

class EndpointOperation(operationBase):
    def perform_operation(self, operation_data: Dict) -> None:
        logging.info("Performing Endpoint Call")
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])

        operation_parameters = operation_data.get("operation_parameters", {})

        endpoint = operation_parameters["endpoint"]
        type = operation_parameters["type"]
        params = operation_parameters["params"]
        jq_filter = operation_parameters["jq"]
        if type == 'GET':
            resp = self.rest_reader.get(endpoint, params=params, filter_query=jq_filter)
        self.investigation_storage.data[target_category][target_items] = resp


class UnionDistinctOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing UnionOperation")

        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})

        distinct_keys = operation_parameters["primary_parameters"]
        key_exists = self.check_items_exist_across_all_files(
            source_category, distinct_keys
        )
        if not key_exists:
            distinct_keys = operation_parameters["secondary_parameters"]
        key_exists = self.check_items_exist_across_all_files(
            source_category, distinct_keys
        )
        if not key_exists:
            raise ItemDoNotExist

        collated_data = self.reader.collate_items(source_category, source_items)
        reduced_collated_data = self.reduce_rows_by_distinct_key(
            collated_data, distinct_keys
        )
        if target_items != "_same":
            for index, item in enumerate(target_items):
                self.rename_item(reduced_collated_data, source_items[index], item)
        self.investigation_storage.set_items(target_category, reduced_collated_data)


class AutoIncrementOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing AutoIncrementOperation")
        target_category = operation_data.get("target_category")
        target_item = operation_data.get("target_item")

        data = self.investigation_storage.data[target_category]
        rows = self.get_number_of_rows_in_data(data)
        self.investigation_storage.data[target_category][target_item] = []
        for i in range(rows):
            data[target_item].append(i + 1)


class StaticValueOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing StaticValue Operation")
        target_category = operation_data.get("target_category")
        target_items = operation_data.get("target_items")
        target_values = operation_data.get("target_values")

        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]
        rows_to_write = max(self.get_number_of_rows_in_data(data), 1)

        for index, item in enumerate(target_items):
            if item not in data:
                data[item] = []
            for i in range(rows_to_write):
                data[item].append(target_values[index])


class ModifyOperation(operationBase):
    # TODO: Change operation name to ModifyIntersection

    def perform_operation(self, operation_data):
        logging.info("Performing Union Modify Operation")
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", "")

        collated_data = self.reader.collate_item(source_category, source_items[0])
        same_data_across_files = collated_data.count(collated_data[0]) == len(
            collated_data
        )
        if not same_data_across_files:
            logging.info(
                "Values are not same across the files, Intersection Operation could not be performed"
            )
            # TODO: put a quesiton mark for the value:
        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]
        for index, item in enumerate(target_items):
            data[item] = [operation_parameters.format(collated_data[0])]


class IntersectionOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing IntersectionOperation")
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])

        collated_data = self.reader.collate_item(source_category, source_items[0])
        same_data_across_files = collated_data.count(collated_data[0]) == len(
            collated_data
        )
        if not same_data_across_files:
            logging.warning(
                "Values are not same across the files, Intersection Operation could not be performed"
            )
            # Put a "?" here as value
        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]
        for index, item in enumerate(target_items):
            data[item] = [collated_data[0]]


class ConditionalUnionOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing ConditionalUnion Operation")
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})

        conditional_tag = operation_parameters["conditional_variable"]
        conditional_values = operation_parameters["value"]

        conditional_item = conditional_tag.split(".")[-1]
        collation_items = deepcopy(source_items)
        if conditional_item not in source_items:
            collation_items.append(conditional_item)

        collated_data = self.reader.collate_items(source_category, collation_items)
        reduced_collated_data = self.reduce_rows_by_condition(
            collated_data, conditional_tag, conditional_values
        )

        for i in range(len(source_items)):
            reduced_collated_data = self.rename_item(
                reduced_collated_data, source_items[i], target_items[i]
            )

        filtered_data = self.filter_rows_by_items(reduced_collated_data, target_items)
        self.investigation_storage.set_items(target_category, filtered_data)


class ConditionalDistinctUnionOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing ConditionalUnion Operation")
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})

        conditional_tag = operation_parameters["conditional_variable"]
        conditional_values = operation_parameters["value"]

        conditional_item = conditional_tag.split(".")[-1]
        collation_items = deepcopy(source_items)
        if conditional_item not in source_items:
            collation_items.append(conditional_item)

        collated_data = self.reader.collate_items(source_category, collation_items)
        reduced_collated_data = self.reduce_rows_by_condition(
            collated_data, conditional_tag, conditional_values
        )
        reduced_collated_data = self.reduce_rows_by_distinct_key(
            reduced_collated_data, operation_parameters["distinct_key"]
        )

        for i in range(len(source_items)):
            reduced_collated_data = self.rename_item(
                reduced_collated_data, source_items[i], target_items[i]
            )

        filtered_data = self.filter_rows_by_items(reduced_collated_data, target_items)
        self.investigation_storage.set_items(target_category, filtered_data)


class CopyFillOperation(operationBase):
    # In copy, the data is copied to and from investigation file itself.
    # It copies the element of the first row of the source,
    # Writes it as many times as the target rows.

    def perform_operation(self, operation_data):
        logging.info("Performing CopyFill Operation")
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category")
        target_items = operation_data.get("target_items")

        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]
        rows_to_write = max(self.get_number_of_rows_in_data(data), 1)

        for index, item in enumerate(target_items):
            source_data = self.investigation_storage.get_item_data(
                source_category, source_items[index]
            )
            if item not in data:
                data[item] = []
            for i in range(rows_to_write):
                data[item].append(source_data[0])


class CopyConditionalModificationOperation(operationBase):
    # In copy, the data is copied to and from investigation file itself.
    # It checks for condition for each row of data with corresponding row of conditional variable
    # If value matches, copy the modified string
    # Else writes "?"

    def perform_operation(self, operation_data):
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category")
        target_items = operation_data.get("target_items")
        operation_parameters = operation_data.get("operation_parameters", {})
        value_to_check = operation_parameters["value"]
        modificaiton = operation_parameters["modification"]

        conditional_tag = operation_parameters["conditional_variable"]
        conditional_category, conditional_item = conditional_tag.split(".")

        data = self.investigation_storage.data[target_category]
        self.investigation_storage.add_category(target_category)

        for index, item in enumerate(target_items):
            source_data = self.investigation_storage.get_item_data(
                source_category, source_items[index]
            )
            conditional_data = self.investigation_storage.get_item_data(
                conditional_category, conditional_item
            )
            if item not in data:
                data[item] = []
            for index, row in enumerate(source_data):
                if conditional_data[index] == value_to_check:
                    data[item].append(modificaiton.format(row))
                else:
                    data[item].append("?")


class CopyOperation(operationBase):
    # In copy, the data is copied to and from investigation file itself.

    def perform_operation(self, operation_data):
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category")
        target_items = operation_data.get("target_items")

        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]

        for index, item in enumerate(target_items):
            source_data = self.investigation_storage.get_item_data(
                source_category, source_items[index]
            )
            data[item] = source_data


class CopyForEachRowOperation(operationBase):
    """
    Source is the model file unless specified
    Target is the investigation file

    It will check  value in  from the investigation file
    And find appropriate row it exists in the model file
    it will copy the source item at that index to the target.
    """

    def perform_operation(self, operation_data):
        logging.info("Performing CopyForEachRow Operation")
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})

        conditional_tag = operation_parameters["conditional_variable"]
        conditional_values = operation_parameters["value"]
        copy_from_investigation = operation_parameters.get(
            "source_is_investigation", ""
        )

        conditional_cat, conditional_item = conditional_tag.split(".")
        collation_items = deepcopy(source_items)

        if copy_from_investigation:
            source_data = self.investigation_storage.get_items_data(
                source_category, source_items
            )
        else:
            source_data = self.reader.collate_items(source_category, collation_items)

        repeats = operation_parameters.get("repeats", False)
        if copy_from_investigation:
            conditional_data = list(self.investigation_storage.get_items_data(
                conditional_cat, [conditional_item]
            ).values())[0]
        elif repeats:
            conditional_data = self.reader.collate_item_per_file(
                conditional_cat, conditional_item
            )
        else:
            conditional_data = self.reader.collate_item(
                conditional_cat, conditional_item
            )
        row_values = self.evaluate_variable(conditional_values)
        result = {}
        field_for_file_no = operation_parameters.get("model_file_id", False)
        if field_for_file_no:
            result.setdefault(field_for_file_no, [])

        for conditional_index, value in enumerate(row_values):
            for source_item in source_items:
                result.setdefault(source_item, [])
                if repeats:
                    for file_index, conditional in conditional_data.items():
                        if copy_from_investigation:
                            if value in conditional:
                                result[source_item].append(conditional_index + 1)
                                if field_for_file_no:
                                    result[field_for_file_no].append(file_index + 1)
                        else:
                            indices_of_value = [
                                i for i, x in enumerate(conditional) if x == value
                            ]
                            for index_of_value in indices_of_value:
                                result[source_item].append(
                                    source_data[source_item][index_of_value]
                                )
                                if field_for_file_no:
                                    result[field_for_file_no].append(file_index + 1)
                else:
                    index_of_value = conditional_data.index(value)
                    result[source_item].append(source_data[source_item][index_of_value])

        for i in range(len(source_items)):
            result = self.rename_item(result, source_items[i], target_items[i])

        if field_for_file_no:
            target_items.insert(0, field_for_file_no)
        filtered_data = self.filter_rows_by_items(result, target_items)
        self.investigation_storage.set_items(target_category, filtered_data)


class DeletionOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing Deletion Operation")
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])

        for item in target_items:
            self.investigation_storage.data[target_category].pop(item, None)


class ExternalInformationOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing External Information Operation")
        # Source category is from the investigation file.
        source_category = operation_data.get("source_category", "")
        source_items = operation_data.get("source_items", [])
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})
        filename = operation_parameters.get("file")

        external_inchi = ExternalInformation("./external_data/" + filename)
        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category].setdefault(
            target_items[0], []
        )

        items = self.investigation_storage.get_item_data(
            source_category, source_items[0]
        )
        for item in items:
            key = external_inchi.get_inchi_key(item)
            items = self.investigation_storage.data[target_category][
                target_items[0]
            ].append(key)


class SQLOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing SQL query Operation")
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})
        query = operation_parameters.get("query", "")

        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]

        response = self.reader.sql_execute(query)
        for result in response:
            for index, item in enumerate(target_items):
                data.setdefault(item, [])
                data[item].append(result[index])

class CopyFromPickleOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing CopyFromPickle Operation")
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        source_items = operation_data.get("source_items", [])

        self.investigation_storage.add_category(target_category)
        for index, item in enumerate(source_items):
            self.investigation_storage.data[target_category].setdefault(target_items[index], [])
            # Issues: semi-colons seperated multiple values. e.g.
            # 'primary_citation_author_name' ='Aschenbrenner, J.C.;Fairhead, M.;Godoy, A.S.;Balcomb, B.H.;Capkin, E.;Chandran, A.V.;Dolci, I.;Golding, M.;Koekemoer, L.;Lithgo, R.M.;
            
            if isinstance(self.reader[item], list):
                self.investigation_storage.data[target_category][target_items[index]].extend(self.reader[item])
            else:
                self.investigation_storage.data[target_category][target_items[index]].append(self.reader[item])

class NoopOperation(operationBase):
    def perform_operation(*args, **kwargs):
        pass

class JQFilterOperation(operationBase):
    def perform_operation(self, operation_data):
        logging.info("Performing JQ Filter Operation")
        target_category = operation_data.get("target_category", "")
        target_items = operation_data.get("target_items", [])
        operation_parameters = operation_data.get("operation_parameters", {})
        jq_filter = operation_parameters.get("jq", "")
        logging.info(f"Category: {target_category}, Item(s): {target_items}, JQ Filter: {jq_filter}")

        # Get filtered data from JSON reader
        filtered_data = self.reader.jq_filter(jq_filter)
        
        self.investigation_storage.add_category(target_category)
        data = self.investigation_storage.data[target_category]

        # Handle single target item case
        if isinstance(target_items, str):
            target_items = [target_items]
            if not isinstance(filtered_data, list):
                filtered_data = [filtered_data]
            data[target_items[0]] = filtered_data
            return

        # Handle multiple target items case
        if len(target_items) != len(filtered_data):
            raise ValueError(f"Number of target items ({len(target_items)}) does not match filtered data length ({len(filtered_data)})")
            
        for item, value in zip(target_items, filtered_data):
            if isinstance(value, list):
                data[item] = value
            else:
                data[item] = [value]



from setuptools import setup, find_packages
import os

# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="mmcif_gen",
    version="1.0.4",
    packages=find_packages(),
    install_requires=[
        "argparse",
        "gemmi",
        "requests",
        "jq",
    ],
    entry_points={
        "console_scripts": [
            "mmcif-gen=mmcif_gen.mmcif_gen:main",
        ],
    },
    author="Syed Ahsan Tanweer",
    author_email="ahsan@ebi.ac.uk",
    description="CLI tool for creating mmCIF files from various facility data sources",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PDBeurope/Investigations/",
    project_urls={
        "Bug Tracker": "https://github.com/PDBeurope/Investigations/issues",
        "Documentation": "https://github.com/PDBeurope/Investigations/",
        "Source Code": "https://github.com/PDBeurope/Investigations/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    python_requires=">=3.6",
    keywords="mmcif, crystallography, structural-biology, pdbe, synchrotron",
)